#[cfg(feature = "bytemuck")]
pub mod impl_bytemuck;

pub mod color;

mod camera;
mod geometry;

use core::f32;

pub use camera::Camera;
pub use geometry::{Point, Transform, Vector, shapes::Shape};

use color::Rgba;
use geometry::{
    Ray,
    shapes::{Hit, Hittable},
};

pub struct Scene {
    camera: Camera,
    contents: Vec<Shape>,
}

impl Scene {
    pub fn new(camera: Camera, contents: Vec<Shape>) -> Self {
        Self { camera, contents }
    }

    pub fn render(&self, width: usize, height: usize) -> Box<[Rgba]> {
        let ray = self.camera.ray(0.0, 0.0);
        dbg!(&ray);
        dbg!(self.intersect(ray));
        let mut data = Box::new_uninit_slice(width * height);
        let w = width as f32;
        let h = height as f32;
        let aspect_ratio = h / w;

        for (i, pixel) in data.iter_mut().enumerate() {
            let x = ((i % width) as f32 * 2.0 / w) - 1.0;
            let y = (1.0 - (i / width) as f32 * 2.0 / h) * aspect_ratio;

            let rgb = self
                .intersect(self.camera.ray(x, y))
                .map(|h| h.normal.normalize() * 0.5 + Vector::new(0.5, 0.5, 0.5))
                .unwrap_or(Vector::ZERO);

            pixel.write(Rgba::new(rgb.x, rgb.y, rgb.z, 1.0));
        }

        unsafe { data.assume_init() }
    }

    fn intersect(&self, ray: Ray) -> Option<Hit> {
        let mut hit: Option<Hit> = None;
        for shape in &self.contents {
            let t_max = hit.as_ref().map(|h| h.t).unwrap_or(f32::INFINITY);
            if let Some(h) = shape.intersect(&ray, t_max) {
                hit = Some(h);
            }
        }
        hit
    }
}
